#include "tp.h"
#include "list.c"
#include "dict.c"
#include "util.c"
#include "string.c"
#include "builtins.c"


tp_obj tp_str(tp_obj self) {
    int type = self.type;
    if (type == TP_STRING) { return self; }
    tp_obj r; char *s;
    r.type = TP_STRING;
    if (type == TP_NUMBER) {
        r.string.buf = _tp_number2char(self.number.value);
    } else if(type == TP_DICT) {
        r.string.buf = tp_printf("<dict 0x%x>",self.dict.value);
    } else if(type == TP_LIST) {
        r.string.buf = tp_printf("<list 0x%x>",self.list.value);
    } else if(type == TP_PARAMS) {
        r.string.buf = tp_printf("<params 0x%x>",self.list.value);
    } else if (type == TP_NONE) {
        r.string.buf = "None";
    } else if (type == TP_DATA) {
        r.string.buf = tp_printf("<data 0x%x>",self.data.value);
    } else if (type == TP_FNC) {
        r.string.buf = tp_printf("<fnc 0x%x>",self.fnc.value);
    } else if (type == TP_DEF) {
        r.string.buf = tp_printf("<def 0x%x>",self.def.value);
    } else {
        r.string.buf = "<?>";
    }
    r.string.length = strlen(r.string.buf);
    return r;
}


tp_obj tp_has(tp_obj self, tp_obj k) {
    int type = self.type;
    if (type == TP_PARAMS && self.list.value->length > 0) { return True; }
    if (type == TP_DICT || type == TP_PARAMS) {
        if (_tp_dict_get(self.dict.value,tp_cstring(tp_str(k)),0)) { return True; }
        return False;
    } else if (type == TP_STRING && k.type == TP_STRING) {
        char *p = strstr(tp_cstring(self),tp_cstring(k));
        return tp_number(p != 0);
    } else if (type == TP_LIST) {
        return tp_number(_tp_list_find(self.list.value,k)!=-1);
    }
    vm_raise("tp_has(%s,%s)",STR(self),STR(k));
}

tp_obj tp_del(tp_obj self, tp_obj k) {
    int type = self.type;
    if (type == TP_DICT) {
        _tp_dict_del(self.dict.value,tp_cstring(tp_str(k)),"tp_del");
        return None;
    }
    vm_raise("tp_del(%s,%s)",STR(self),STR(k));
}

tp_obj tp_iter(tp_obj self, tp_obj k) {
    int type = self.type;
    if (type == TP_LIST || type == TP_STRING) { return tp_get(self,k); }
    if (type == TP_DICT && k.type == TP_NUMBER) {
        return tp_string(self.dict.value->items[(int)k.number.value].key);
    }
    vm_raise("tp_iter(%s,%s)",STR(self),STR(k));
}


tp_obj tp_get(tp_obj self, tp_obj k) {
    int type = self.type;
    if (type == TP_DICT) {
        return *(tp_obj*)_tp_dict_get(self.dict.value,tp_cstring(tp_str(k)),"tp_get");
    } else if (type == TP_LIST) {
        if (k.type == TP_NUMBER) {
            int l = tp_len(self).number.value;
            int n = k.number.value;
            n = (n<0?l+n:n);
            return *(tp_obj*)_tp_list_get(self.list.value,n,"tp_get");
        } else if (k.type == TP_STRING) {
            if (strcmp("append",tp_cstring(k)) == 0) {
                return tp_method(self,append);
            } else if (strcmp("pop",tp_cstring(k)) == 0) {
                return tp_method(self,tp_pop);
            } else if (strcmp("index",tp_cstring(k)) == 0) {
                return tp_method(self,tp_index);
            } else if (strcmp("sort",tp_cstring(k)) == 0) {
                return tp_method(self,tp_sort);
            } else if (strcmp("extend",tp_cstring(k)) == 0) {
                return tp_method(self,extend);
            }
        }
    } else if (type == TP_STRING) {
        if (k.type == TP_NUMBER) {
            int l = self.string.length;
            int n = k.number.value;
            n = (n<0?l+n:n);
            if (n < 0 || n >= l) {
                vm_raise("tp_get(%s,%s)",STR(self),STR(k));
            }
            return tp_string_n(tp_printf("%c",self.string.buf[n]),1);
        } else if (k.type == TP_STRING) {
            if (strcmp("join",tp_cstring(k)) == 0) {
                return tp_method(self,join);
            } else if (strcmp("split",tp_cstring(k)) == 0) {
                return tp_method(self,split);
            } else if (strcmp("index",tp_cstring(k)) == 0) {
                return tp_method(self,tp_str_index);
            } else if (strcmp("strip",tp_cstring(k)) == 0) {
                return tp_method(self,tp_strip);
            }
        }
    } else if (type == TP_PARAMS) {
        if (k.type == TP_STRING) {
            if (k.string.buf[0] == '[') {
                tp_obj r; r.type = TP_LIST;
                r.list.value = _tp_list_copy(self.list.value);
                self.list.value->length = 0;
                return r;
            } else if (k.string.buf[0] == '{') {
                tp_obj r; r.type = TP_DICT;
                r.dict.value = _tp_dict_copy(self.dict.value);
                self.dict.value->length = 0;
                return r;
            } else if (self.list.value->length) {
                return *(tp_obj*)_tp_list_pop(self.list.value,0,"tp_get");
            } else {
                return *(tp_obj*)_tp_dict_get(self.dict.value,tp_cstring(tp_str(k)),"tp_get");
            }
        }
    } else if (type == TP_DATA) {
        return _call2(self.data.meta[M_GET],self,k);
    }

    if (k.type == TP_LIST) {
        int a,b,l,i;
        tp_obj tmp;
        l = tp_len(self).number.value;
        tmp = tp_get(k,tp_number(0));
        if (tmp.type == TP_NUMBER) { a = tmp.number.value; }
        else if(tmp.type == TP_NONE) { a = 0; }
        else { vm_raise("%s is not a number",STR(tmp)); }
        tmp = tp_get(k,tp_number(1));
        if (tmp.type == TP_NUMBER) { b = tmp.number.value; }
        else if(tmp.type == TP_NONE) { b = l; }
        else { vm_raise("%s is not a number",STR(tmp)); }
        a = (a<0?l+a:a); b = (b<0?l+b:b);
        if (type == TP_LIST) {
            tp_obj r = tp_list(0,0);
            for (i=a; i<b; i++) {
                _tp_list_append(r.list.value,tp_ptr(tp_get(self,tp_number(i))));
            }
            return r;
        } else if (type == TP_STRING) {
            char *ptr = tp_malloc(b-a+1);
            memcpy(ptr,self.string.buf+a,b-a); ptr[b-a]=0;
            return tp_string_n(ptr,b-a);
        }
    }
    vm_raise("tp_get(%s,%s)",STR(self),STR(k));
}

tp_obj tp_dget(tp_obj self, tp_obj k, tp_obj d) {
    if (tp_has(self,k).number.value) { return tp_get(self,k); }
    return d;
}
tp_obj tp_iget(tp_obj *r, tp_obj self, tp_obj k) {
    if (tp_has(self,k).number.value) { *r = tp_get(self,k); }
    return;
}

void tp_set(tp_obj self, tp_obj k, tp_obj v) {
    int type = self.type;
    if (type == TP_DICT) {
        _tp_dict_set(self.dict.value,tp_cstring(tp_str(k)),tp_ptr(v));
        return;
    } else if (type == TP_LIST) {
        _tp_list_set(self.list.value,k.number.value,tp_ptr(v),"tp_set");
        return;
} else if (type == TP_PARAMS && k.type == TP_STRING) {
        if (k.string.buf[0] == '[') {
            extend(tp_params(2,(tp_obj[]){self,v}));
        } else if (k.string.buf[0] == '{') {
            merge(tp_params(2,(tp_obj[]){self,v}));
        } else {
            _tp_dict_set(self.dict.value,tp_cstring(tp_str(k)),tp_ptr(v));
        }
        return;
    } else if (type == TP_DATA) {
        _call3(self.data.meta[M_SET],self,k,v);
        return;
    }
    vm_raise("tp_set(%s,%s,%s)",STR(self),STR(k),STR(v));
}

tp_obj tp_add(tp_obj a, tp_obj b) {
    if (a.type == TP_NUMBER && a.type == b.type) {
        return tp_number(a.number.value+b.number.value);
    } else if (a.type == TP_STRING && a.type == b.type) {
        int al = a.string.length, bl = b.string.length;
        char *s = tp_malloc(al+bl+1); s[al+bl] = 0;
        memcpy(s,a.string.buf,al); memcpy(s+al,b.string.buf,bl);
        return tp_string_n(s,al+bl);
    } else if (a.type == TP_LIST && a.type == b.type) {
        tp_obj r = tp_copy(tp_params(1,(tp_obj[]){a}));
        extend(tp_params(2,(tp_obj[]){r,b}));
        return r;
    }
    vm_raise("tp_add(%s,%s)",STR(a),STR(b));
}


tp_obj tp_mul(tp_obj a, tp_obj b) {
    if (a.type == TP_NUMBER && a.type == b.type) {
        return tp_number(a.number.value*b.number.value);
    } else if (a.type == TP_STRING && b.type == TP_NUMBER) {
        int al = a.string.length; int n = b.number.value;
        char *s = tp_malloc(al*n+1); s[al*n] = 0;
        int i; for (i=0; i<n; i++) { memcpy(s+al*i,a.string.buf,al); }
        return tp_string_n(s,al*n);
    }
    vm_raise("tp_mul(%s,%s)",STR(a),STR(b));
}


tp_obj tp_len(tp_obj self) {
    int type = self.type;
    if (type == TP_STRING) {
        return tp_number(self.string.length);
    } else if (type == TP_DICT) {
        return tp_number(self.dict.value->length);
    } else if (type == TP_LIST) {
        return tp_number(self.list.value->length);
    }
    vm_raise("tp_len(%s)",STR(self));
}

tp_obj tp_cmp(tp_obj a, tp_obj b) {
    int type = a.type;
    if (type != b.type) {
        return tp_number(a.type-b.type);
    } else if (type == TP_NONE) {
        return tp_number(0);
    } else if (type == TP_NUMBER) {
        return tp_number(a.number.value-b.number.value);
    } else if (type == TP_STRING) {
        int v = strncmp(a.string.buf,b.string.buf,MIN(a.string.length,b.string.length));
        if (v == 0) { v = a.string.length-b.string.length; }
        return tp_number(v);
    }
    vm_raise("tp_cmp(%s,%s)",STR(a),STR(b));
}

#define TP_OP(name,expr) \
    tp_obj name(tp_obj _a,tp_obj _b) { \
    if (_a.type == TP_NUMBER && _a.type == _b.type) { \
        float a = _a.number.value; float b = _b.number.value; \
        return tp_number(expr); \
    } \
    vm_raise("%s(%s,%s)",#name,STR(_a),STR(_b)); \
}

TP_OP(tp_and,((int)a)&((int)b));
TP_OP(tp_or,((int)a)|((int)b));
TP_OP(tp_mod,((int)a)%((int)b));
TP_OP(tp_lsh,((int)a)<<((int)b));
TP_OP(tp_rsh,((int)a)>>((int)b));
TP_OP(tp_sub,a-b);
TP_OP(tp_div,a/b);
TP_OP(tp_pow,pow(a,b));

//
