#ifndef TP_H
#define TP_H

#include <setjmp.h>
#include <sys/stat.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <stdint.h>

#include <gc/gc.h>
#define tp_malloc GC_MALLOC
#define tp_realloc GC_REALLOC

union tp_obj;
typedef int tint;
typedef enum tp_types {
    TP_NONE,TP_NUMBER,TP_STRING,TP_DICT,
    TP_LIST,TP_PARAMS,TP_FNC,TP_DEF,TP_DATA,
} tp_types;

typedef enum tp_metas {
    M_GET,M_SET,M_STR,M_LEN, M_HAS,M_DEL,M_TOTAL,
} tp_metas;

typedef struct tp_number_ {
    tint type;
    float value;
} tp_number_;

typedef struct tp_string_ {
    tint type;
    char *buf;
    int length;
} tp_string_;

typedef struct tp_item {
    char *key;
    void *value;
} tp_item;
typedef struct _tp_dict {
    tp_item *items;
    int length;
    int alloc;
} _tp_dict;
typedef struct _tp_list {
    void **items;
    int length;
    int alloc;
} _tp_list;

typedef struct tp_list_ {
    tint type;
    _tp_list *value;
    void *_;
} tp_list_;
typedef struct tp_dict_ {
    tint type;
    void *_;
    _tp_dict *value;
} tp_dict_;
typedef struct tp_params_ {
    tint type;
    _tp_list *list;
    _tp_dict *dict;
} tp_params_;


typedef struct tp_fnc_ {
    tint type;
    void *value;
    void *self;
    void *globals;
} tp_fnc_;

typedef struct tp_data_ {
    tint type;
    void *value;
    void **meta;
} tp_data_;

typedef union tp_obj {
    tint type;
    tp_number_ number;
    tp_string_ string;
    tp_dict_ dict;
    tp_list_ list;
    tp_params_ params;
    tp_fnc_ fnc;
    tp_fnc_ def;
    tp_data_ data;
} tp_obj;

#define vm_raise(fmt,...) _vm_raise(tp_string(tp_printf(fmt,__VA_ARGS__)))
#define STR(v) (tp_cstring(tp_str(v)))

int MIN(int a, int b) { return (a<b?a:b); }
int MAX(int a, int b) { return (a>b?a:b); }

#define None ((tp_obj){TP_NONE})
#define True tp_number(1)
#define False tp_number(0)

#define P_BEGIN(params) tp_obj __params = (params)
#define P_OBJ(k) (tp_get(__params,tp_string(k)))
#define P_POP() *(tp_obj*)_tp_list_pop(__params.list.value,0,"P_POP")
#define P_NUMBER(k) (P_OBJ(k).number.value)
#define P_STRING(k) (tp_cstring(P_OBJ(k)))
#define P_DEFAULT(k,d) (tp_dget(__params,tp_string(k),d))
#define P_END
#define P_LOOP_BEGIN(e) \
    int __l = __params.list.value->length; \
    int __i; for (__i=0; __i<__l; __i++) { \
    (e) = *(tp_obj*)_tp_list_get(__params.list.value,__i,"P_LOOP");
#define P_LOOP_END \
    }


tp_obj tp_string(char *value);
tp_obj tp_string_n(char *value,int );
inline tp_obj tp_number(float v) { return (tp_obj)(tp_number_){TP_NUMBER,v}; }

void tp_set(tp_obj,tp_obj,tp_obj);
tp_obj tp_get(tp_obj,tp_obj);
tp_obj tp_len(tp_obj);
tp_obj tp_str(tp_obj);
tp_obj *tp_ptr(tp_obj o);
tp_obj tp_cmp(tp_obj,tp_obj);
tp_obj tp_dget(tp_obj,tp_obj,tp_obj);

void _vm_raise(tp_obj);
char *tp_printf(char *fmt,...);
char *tp_cstring(tp_obj);

#endif
//
